// Mega 1284 p


#define F_CPU 16000000UL                
#include "trtSettings.h"
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdio.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <stdint.h>

#include "trtkernel_1284.c"
// serial communication library
// Don't mess with the semaphores
#define FXNSTACKSIZE 350
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#define SEM_SHARED 3
#define SEM_LCD 4
#define SEM_XACCEL 5
#define SEM_STRIDE 6
#define Calibration1G 341.0 //
#define sliding_window 4

//LCD stuff
#include "lcd_lib.h"
uint8_t lcd_char[17];       //buffer to write to LCD, int long is 32 bits in avr-gcc
uint8_t lcd_char2[17];
#include "trtUart.h"
#include "trtUart.c"

// UART file descriptor
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
// input arguments to each thread, not actually used in this example
int args[3];

void initialize(void);                 //all the usual mcu stuff
void task1_configuration(void* args);  //configure stride and pace
void task2_update_LCD(void* args);     //calculate velocity and update LCD
void task3_recv_bluetooth(void* args); //Receive packets from bluetooth

volatile long sum_accel = 0;
volatile int sum_accel_overflow = 0;
volatile unsigned long timestamp_step = 0; //delta t between summations
volatile float desired_speed = 0;
volatile uint8_t configmode = 1;
volatile unsigned int num_step = 0;
volatile float stride_length = 0;
volatile unsigned long timestamp_array[sliding_window] = {0,0,0,0};

//***************************************************************************
// Task 3 handles communication with the foot unit via Bluetooth/UART link
void task3_recv_bluetooth(void* args){
  unsigned long timestamp = 0;
  unsigned long timestamp_last_peak = 0;
  unsigned long timestamp_diff;

  int accel = 42;
  int accel_old = 0;
  int axis; 
  int jerk = -1; //prevents initial incorrent pos to negative jerk which indicates accel max
  int jerk_old;
  long sum;

  int i;

  int accel_thresh = 869;   //max is 1023

  while(1){
    // Read packet data 
    fscanf(stdin, "%d,%d,%lu,%ld", &axis, &accel, &timestamp, &sum);

    //Read z axis packet data: [axis,accel,timestamp,0]
    if (axis == 0){
      //update jerk variables
      jerk_old = jerk;
      jerk = accel - accel_old;
      
      //If the jerk goes from positive to negative, we had an accel peak
      if (jerk_old >= 0 && jerk < 0){
  
        //look at acceleration and make sure it's large enough to be an actual peak
        if (accel > accel_thresh || (jerk > 20 && accel > 750)){
          //"Debounce" the step

          //Overflow case
          if (timestamp < timestamp_last_peak){
            timestamp_diff = timestamp + (ULONG_MAX - timestamp_last_peak);
          }
          else {
            timestamp_diff = timestamp - timestamp_last_peak;
          }
          
          //Make sure that peaks are far enough away to make sense
          //if so, step has been detected
          if (timestamp_diff > 3200){ //~200ms
            timestamp_last_peak = timestamp;
            PORTA ^= 0x04; //toggle the red led for a step
            fprintf(stdout, "%d\n", 1); //bluetooth a one for a step

            //capture time of step for pace calculation
            trtWait(SEM_STRIDE);
            //shift in the newest timestamp data each time there's a step
            for (i = sliding_window -1; i > 0; i--){
              timestamp_array[i] = timestamp_array[i-1];
            }
            timestamp_array[0] = timestamp; //would have more accurate data 
            trtSignal(SEM_STRIDE);

          }
          
        }
      }
      accel_old = accel;
      PORTA ^= 0x08;    //toggle LED every time something is received
    }

    //Read x axis packet data: [axis,sum_accel_overflow,delta(timestamp),sum_accel]
    else if (axis == 1){
      //toggle LED for successful communication with the foot
      PORTA ^= 0x02;

      //update global variables for task2
      trtWait(SEM_XACCEL);
      num_step++;
      sum_accel = sum;
      sum_accel_overflow = accel;
      timestamp_step = timestamp;
      trtSignal(SEM_XACCEL);     
    }
  }

}

//***********************************************
//  Calculate the pace and update the LCD
void task2_update_LCD(void* args) {
  uint32_t rel,dead;
  int str_len;
  int i;
  float avgvelocity;
  float avgvelocity_last_valid = 42.0; 
  uint8_t prev_switch = 0x00;
  uint8_t minutes;
  uint8_t seconds;


  //avg veloctiy vars
  
  float local_stride_length;

  while (1) {
    PORTA ^= 0x01;
    
    //if switch has been toggled (only works for falling edge so that config mode isn't constantly set)
    if((PINA & 0x10) == 0x00 && prev_switch == 0x10){ 
      configmode = 1;
    }
    prev_switch = PINA & 0x10;

    //if not config resume normal LCD calculations
    if (configmode == 0){

      //SEM_STRIDE is protecting timestamp_array set in task1 and task3
      trtWait(SEM_STRIDE);
      local_stride_length = stride_length;
      if ( timestamp_array[0] > timestamp_array[sliding_window -1] && ((timestamp_array[0] - timestamp_array[sliding_window -1]) != 0)){
        avgvelocity = local_stride_length * 4.0 / (float)(timestamp_array[0] - timestamp_array[sliding_window -1]);
      }
      else{
        avgvelocity = avgvelocity_last_valid;
      }

      //convert from m/tick to minutes/mile
      if (avgvelocity != 0){
          avgvelocity = 26.817/((float)TICKSPERSECOND * avgvelocity);
          avgvelocity_last_valid = avgvelocity;
      }
      else {
        avgvelocity = avgvelocity_last_valid;
      }
      trtSignal(SEM_STRIDE);

      //convert from m/s to min per mile
      if((avgvelocity > -0.001) && (avgvelocity < 0.001)){ //can't directly == floats
        minutes = 0;
        seconds = 0;
      }
      else{
        avgvelocity = avgvelocity*60.0;
        minutes = ((int)avgvelocity)/60;
        seconds = ((int)avgvelocity)%60;
      }

      //convert to string for display
      if(seconds >= 10){
        snprintf(lcd_char, 17, "Pace: %d:%d %d", minutes, seconds, num_step);
      }
      else{
        if(seconds == 0){
          snprintf(lcd_char, 17, "Pace: %d:%d0 %d", minutes, seconds, num_step);
        }
        else{
          snprintf(lcd_char, 17, "Pace: %d:0%d %d", minutes, seconds, num_step);
        }
      }

      //clear row if necessary
      str_len = strlen(lcd_char);
      for(i = str_len; i < 17; i ++){
        lcd_char[i] = ' ';
      }

      //SEM_LCD protects LCD code
      trtWait(SEM_LCD);
      LCDGotoXY(0,0);
      LCDstring(lcd_char, strlen(lcd_char));
      trtSignal(SEM_LCD);
    }
    //Enter into Config mode!
    else{
      trtSignal(SEM_SHARED);
    }

    rel = trtCurrentTime() + SECONDS2TICKS(.3);
    dead = trtCurrentTime() + SECONDS2TICKS(.4);
    trtSleepUntil(rel, dead);
  }
}

//********************************************
// Task 1 handles configuration and user input
void task1_configuration(void* args) {
  unsigned int min = 15;
  unsigned int tensecond = 0;
  uint8_t buttons; //only care about upper 4 bits of PINA
  uint8_t prevbuttons;
  unsigned int str_len;
  float local_stride = 1.0;
  uint8_t waitforunpressenter = 0;
  while (1) {
    PORTA ^= 0x01;
    //go to sleep until signalled by task 2
    trtWait(SEM_SHARED);
    buttons = PINA & 0xF0;
    while(configmode != 0){
      prevbuttons = buttons;
      //sleep for 50 ms for debounce, don't update deadline b/c we want this done asap
      _delay_ms(50);
      buttons = PINA & 0xF0;
      //if enter button has been previously pushed, wait until unpressed before allowing a button push to be evaluated
      //otherwise we just zip right thought the enter button press code
      if (waitforunpressenter == 1 && (buttons & 0x80) == 0x80){
        waitforunpressenter = 0;
      }

      //Enter button press
      if ((buttons & 0x80) == 0x00 && (prevbuttons & 0x80) == 0x00 && waitforunpressenter == 0){
        if(configmode == 1){      //stride
          snprintf(lcd_char2, 17, "Stride: %.2f", local_stride);
          //protect the global stride_length
          trtWait(SEM_STRIDE);
          stride_length = local_stride;
          trtSignal(SEM_STRIDE);
          configmode = 2;
          waitforunpressenter = 1;
        }
        else{                     //pace
          snprintf(lcd_char2, 17, "Goal: %d:%d0", min, tensecond);
          configmode = 0;
        }

        //clear the row if necessary
        str_len = strlen(lcd_char2);
        for(int i = str_len; i < 17; i ++){
            lcd_char2[i] = ' ';
        }
        trtWait(SEM_LCD);
        LCDGotoXY(0,1);
        LCDstring(lcd_char2, strlen(lcd_char2));
        trtSignal(SEM_LCD);
      }
      //Up button press
      else if ((buttons & 0x20) == 0x00 && (prevbuttons & 0x20) == 0x00){
        if(configmode == 1){        //stride
          //limit at 2 meters
          if(local_stride < 2){
            local_stride = local_stride + .01;
          }
        }
        else{                       //pace
          //up button has been pressed, debounced/same after 30 ms
          //increment time by 30 seconds
          //limit at 90 minutes
          if (tensecond + 1 >= 6){
            if (min < 90) {min++;}
            tensecond = 0;
          }
          else{tensecond++;}
        }
      }
      //Down button press
      else if ((buttons & 0x40) == 0x00 && (prevbuttons & 0x40) == 0x00){
        if(configmode == 1){         //stride
          //limit at .01 meters
          if(local_stride > .01){
            local_stride = local_stride - .01;
          }
        }
        else{                        //pace
          //down has been pressed, debounced/same after 30 ms
          //decrement time by 30 seconds
          //limit at 1 minute
          if (tensecond <= 0){
            if (min > 1) {min--;}
            tensecond = 5;
          }
          else{ tensecond--;}
        }
      }

      if(configmode == 1){
        snprintf(lcd_char2, 17, "Stride(m): %.2f", local_stride);
      }
      else{
        snprintf(lcd_char2, 17, "Min/mile: %d:%d0", min, tensecond);
      }
      //clear the row if necessary
      str_len = strlen(lcd_char2);
      for(int i = str_len; i < 17; i ++){
          lcd_char[i] = ' ';
      }
      trtWait(SEM_LCD);
      LCDGotoXY(0,0);
      LCDstring(lcd_char2, strlen(lcd_char2));
      trtSignal(SEM_LCD);
    }
  }
}

//**********************************************************
//Set it all up
void initialize(void)
{ 
  //Initialize the LCD
  LCDinit();
  LCDcursorOFF();
  LCDclr();
  LCDGotoXY(0,0);

  //init the UART
  trt_uart_init();
  stdout = stdin = stderr = &uart_str;
  
  //Initialize the indicator LED's
  DDRA = 0x0F;          //A.0 keepalive LED
                        //A.1 toggle on receive x axis data
                        //A.2 toggle on step
                        //A.3 toggle on receive z axis data
                        //A.4 config/run switch
                        //A.5 up button
                        //A.6 down button
                        //A.7 enter button
  PORTA = 0xF0;         //pullup on for button/switch (active low)
}

//Entry point for wrist
int main(void){
  initialize();

  //start TRT
  trtInitKernel(150); //150 bytes for idle task stack

  // --- create semaphores ----------
  trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; //UART receive ISR semaphore
  trtCreateSemaphore(SEM_STRING_DONE, 0) ; //UART not even being used semaphore
  trtCreateSemaphore(SEM_SHARED, 0) ; //Semaphore for signaling between task1 and task2
  trtCreateSemaphore(SEM_LCD, 1); //Semaphore protecting LCD code
  trtCreateSemaphore(SEM_XACCEL, 1); //Semaphore protecting global variables updated in task3
  trtCreateSemaphore(SEM_STRIDE, 1); //Semaphore protecting timestamp_array and stride_length global variables
  
  trtCreateTask(task1_configuration, FXNSTACKSIZE, SECONDS2TICKS(0.2), SECONDS2TICKS(.4), &(args[0]));
  trtCreateTask(task2_update_LCD, FXNSTACKSIZE, SECONDS2TICKS(.2), SECONDS2TICKS(.3), &(args[0]));
  trtCreateTask(task3_recv_bluetooth, FXNSTACKSIZE, SECONDS2TICKS(0.1), SECONDS2TICKS(0.2), &(args[0]));

  while (1)
  {
    //I feel loopy
  }
}
